'use strict';

const postcss = require('postcss');
const t = require('@babel/types');
const _babelGenerate = require('@babel/generator');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

function _interopNamespaceCompat(e) {
  if (e && typeof e === 'object' && 'default' in e) return e;
  const n = Object.create(null);
  if (e) {
    for (const k in e) {
      n[k] = e[k];
    }
  }
  n.default = e;
  return n;
}

const postcss__default = /*#__PURE__*/_interopDefaultCompat(postcss);
const t__namespace = /*#__PURE__*/_interopNamespaceCompat(t);
const _babelGenerate__default = /*#__PURE__*/_interopDefaultCompat(_babelGenerate);

function isPlainObject(value) {
  if (value === null || typeof value !== "object") {
    return false;
  }
  const prototype = Object.getPrototypeOf(value);
  if (prototype !== null && prototype !== Object.prototype && Object.getPrototypeOf(prototype) !== null) {
    return false;
  }
  if (Symbol.iterator in value) {
    return false;
  }
  if (Symbol.toStringTag in value) {
    return Object.prototype.toString.call(value) === "[object Module]";
  }
  return true;
}

function _defu(baseObject, defaults, namespace = ".", merger) {
  if (!isPlainObject(defaults)) {
    return _defu(baseObject, {}, namespace, merger);
  }
  const object = Object.assign({}, defaults);
  for (const key in baseObject) {
    if (key === "__proto__" || key === "constructor") {
      continue;
    }
    const value = baseObject[key];
    if (value === null || value === void 0) {
      continue;
    }
    if (merger && merger(object, key, value, namespace)) {
      continue;
    }
    if (Array.isArray(value) && Array.isArray(object[key])) {
      object[key] = [...value, ...object[key]];
    } else if (isPlainObject(value) && isPlainObject(object[key])) {
      object[key] = _defu(
        value,
        object[key],
        (namespace ? `${namespace}.` : "") + key.toString(),
        merger
      );
    } else {
      object[key] = value;
    }
  }
  return object;
}
function createDefu(merger) {
  return (...arguments_) => (
    // eslint-disable-next-line unicorn/no-array-reduce
    arguments_.reduce((p, c) => _defu(p, c, "", merger), {})
  );
}
const defu = createDefu();

const markedLayerKey = Symbol("__tw_layer_name__");
const layerNodesKeys = ["base", "components", "utilities"];

const atRulesNameFilter = "twlayer";
const atRulesRenamePlugin = () => {
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-at-rule-rename-plugin",
    Once(root) {
      root.walkAtRules("layer", (rule) => {
        rule.name = atRulesNameFilter;
      });
    }
  };
};
atRulesRenamePlugin.postcss = true;
const markLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-mark-layer-plugin",
    Once(root) {
      root.walkAtRules(atRulesNameFilter, (rule) => {
        if (layerNodesKeys.includes(rule.params)) {
          const layerName = rule.params;
          for (const node of rule.nodes) {
            Object.defineProperty(node, markedLayerKey, {
              value: layerName,
              enumerable: false,
              configurable: true
            });
          }
          rule.parent?.insertBefore(rule, rule.nodes);
          rule.remove();
        }
      });
      if (ctx.options?.outSideLayerCss !== void 0) {
        root.walkRules((rule) => {
          const layerName = rule[markedLayerKey];
          if (!layerName) {
            Object.defineProperty(rule, markedLayerKey, {
              value: ctx.options?.outSideLayerCss,
              enumerable: false,
              configurable: true
            });
          }
        });
      }
    }
  };
};
markLayerPlugin.postcss = true;
const extractLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-extract-layer-plugin",
    OnceExit(root) {
      root.walkRules((rule) => {
        const layerName = rule[markedLayerKey];
        if (layerName && layerNodesKeys.includes(layerName)) {
          ctx.append(layerName, rule);
        }
      });
      if (ctx.options?.interceptors?.css)
        for (const fn of ctx.options.interceptors.css) {
          fn(root, ctx);
        }
    }
  };
};
extractLayerPlugin.postcss = true;

function _interopDefaultCompat$1(e) {
  return e && typeof e === "object" && "default" in e ? e.default : e;
}
const babelGenerate = _interopDefaultCompat$1(_babelGenerate__default);
const expandAST = ["addBase", "addComponents", "addUtilities", "theme"].map((x) => {
  return t__namespace.objectProperty(t__namespace.identifier(x), t__namespace.identifier("_" + x));
});
function expandAPI() {
  return expandAST;
}
function unwrapThemeFunctionArg(str) {
  if (str) {
    const start = str[0];
    const last = str.at(-1);
    if (start === '"' && last === '"' || start === "'" && last === "'") {
      return str.slice(1, -1);
    }
  }
  return str;
}
function makeObjectProperty(key, value) {
  return t__namespace.objectProperty(t__namespace.stringLiteral(key), value);
}
function addSuffix(value, important) {
  if (typeof value === "string") {
    return value + (important ? " !important" : "");
  }
  return value;
}
function makeObjectExpression(nodes) {
  const res = nodes.reduce((acc, _cur) => {
    if (_cur.type === "rule") {
      const cur = _cur;
      const key = cur.selector;
      const value = cur.nodes.reduce((cacc, ccur) => {
        if (ccur.type === "decl") {
          const v = ccur.value;
          const important = ccur.important;
          const prop = ccur.prop;
          const arr = [...v.matchAll(/theme\(([^()]+?)\)/g)];
          if (arr.length > 0) {
            const first = arr[0];
            if (first[0] === v) {
              const ccc = t__namespace.callExpression(t__namespace.identifier("_theme"), [t__namespace.stringLiteral(unwrapThemeFunctionArg(first[1]))]);
              cacc[prop] = important ? t__namespace.binaryExpression("+", ccc, t__namespace.stringLiteral(" !important")) : ccc;
            } else {
              let p = 0;
              const quasis = [];
              const expressions = [];
              for (let i = 0; i < arr.length; i++) {
                const hit = arr[i];
                if (typeof hit.index === "number") {
                  const raw = v.slice(p, hit.index);
                  p = hit.index + hit[0].length;
                  quasis.push(t__namespace.templateElement({ raw }));
                  expressions.push(t__namespace.callExpression(t__namespace.identifier("_theme"), [t__namespace.stringLiteral(unwrapThemeFunctionArg(hit[1]))]));
                  if (i === arr.length - 1) {
                    quasis.push(t__namespace.templateElement({ raw: addSuffix(v.slice(p), important) }));
                  }
                }
              }
              cacc[prop] = t__namespace.templateLiteral(quasis, expressions);
            }
          } else {
            cacc[prop] = t__namespace.stringLiteral(addSuffix(v, important));
          }
        }
        return cacc;
      }, {});
      if (acc[key]) {
        acc[key].push(
          ...Object.entries(value).map(([prop, value2]) => {
            return makeObjectProperty(prop, value2);
          })
        );
      } else {
        acc[key] = Object.entries(value).map(([prop, value2]) => {
          return makeObjectProperty(prop, value2);
        });
      }
    }
    return acc;
  }, {});
  return Object.entries(res).map(([key, value]) => {
    return t__namespace.objectProperty(
      t__namespace.stringLiteral(key),
      t__namespace.objectExpression(
        Object.values(
          value.reduce((acc, cur) => {
            if (cur.key.type === "StringLiteral") {
              acc[cur.key.value] = cur;
            }
            return acc;
          }, {})
        )
      )
    );
  });
}
function getVarName(layer) {
  return `_${layer}Css`;
}
const pluginNameMap = {
  base: "_addBase",
  components: "_addComponents",
  utilities: "_addUtilities"
};
function getFnName(key) {
  return pluginNameMap[key];
}
function getLegacyStatement(key, ctx) {
  return t__namespace.expressionStatement(t__namespace.callExpression(t__namespace.identifier(getFnName(key)), [t__namespace.objectExpression(makeObjectExpression(ctx.getNodes(key)))]));
}
const returnSelfArrowFnAst = t__namespace.arrowFunctionExpression([t__namespace.identifier("x")], t__namespace.identifier("x"));
const declReturnSelfArrowFnName = "returnSelfNoop";
const declReturnSelfArrowFnAst = t__namespace.variableDeclaration("const", [t__namespace.variableDeclarator(t__namespace.identifier(declReturnSelfArrowFnName), returnSelfArrowFnAst)]);
function createGenerator() {
  const callFnId = "_plugin";
  const pluginName = "css2TwPlugin";
  const withOptionsParamsId = "_options";
  const traverseMethodName = "withOptionsWalkCSSRuleObject";
  const requireStatement = t__namespace.variableDeclaration("const", [
    t__namespace.variableDeclarator(t__namespace.identifier(callFnId), t__namespace.callExpression(t__namespace.identifier("require"), [t__namespace.stringLiteral("tailwindcss/plugin")]))
  ]);
  const exportsStatement = t__namespace.expressionStatement(t__namespace.assignmentExpression("=", t__namespace.memberExpression(t__namespace.identifier("module"), t__namespace.identifier("exports")), t__namespace.identifier(pluginName)));
  function getLatestStatement(key, ctx) {
    const varName = getVarName(key);
    const fnName = getFnName(key);
    return [
      t__namespace.variableDeclaration("const", [
        t__namespace.variableDeclarator(
          t__namespace.identifier(varName),
          t__namespace.callExpression(t__namespace.identifier(traverseMethodName), [t__namespace.objectExpression(makeObjectExpression(ctx.getNodes(key))), t__namespace.stringLiteral(key)])
        )
      ]),
      t__namespace.expressionStatement(t__namespace.callExpression(t__namespace.identifier(fnName), [t__namespace.identifier(varName)]))
    ];
  }
  function generate(ctx, opt) {
    function getPluginBody() {
      const statements = ctx.options?.withOptions ? layerNodesKeys.flatMap((layer) => {
        return getLatestStatement(layer, ctx);
      }) : layerNodesKeys.map((layer) => {
        return getLegacyStatement(layer, ctx);
      });
      const innerPluginAst = t__namespace.functionExpression(
        null,
        [t__namespace.objectPattern(expandAPI())],
        t__namespace.blockStatement(statements)
      );
      if (ctx.options?.withOptions) {
        return t__namespace.callExpression(t__namespace.memberExpression(t__namespace.identifier(callFnId), t__namespace.identifier("withOptions")), [
          t__namespace.functionExpression(
            null,
            [t__namespace.assignmentPattern(t__namespace.identifier(withOptionsParamsId), t__namespace.objectExpression([]))],
            t__namespace.blockStatement([
              // const { withOptionsWalkCSSRuleObject = (x)=>x } = _options
              t__namespace.variableDeclaration("const", [
                t__namespace.variableDeclarator(
                  t__namespace.objectPattern([
                    t__namespace.objectProperty(t__namespace.identifier(traverseMethodName), t__namespace.assignmentPattern(t__namespace.identifier(traverseMethodName), t__namespace.identifier(declReturnSelfArrowFnName)))
                  ]),
                  t__namespace.identifier(withOptionsParamsId)
                )
              ]),
              t__namespace.returnStatement(innerPluginAst)
            ])
          ),
          t__namespace.functionExpression(null, [t__namespace.identifier(withOptionsParamsId)], t__namespace.blockStatement([t__namespace.returnStatement(t__namespace.objectExpression([]))]))
        ]);
      }
      return t__namespace.callExpression(t__namespace.identifier(callFnId), [innerPluginAst]);
    }
    const ast = t__namespace.file(
      t__namespace.program([requireStatement, declReturnSelfArrowFnAst, t__namespace.variableDeclaration("const", [t__namespace.variableDeclarator(t__namespace.identifier(pluginName), getPluginBody())]), exportsStatement])
    );
    const res = babelGenerate(ast, opt);
    return res.code;
  }
  return {
    generate
  };
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class BaseContext {
  constructor() {
    __publicField(this, "layersMap");
    __publicField(this, "options");
    __publicField(this, "generator");
    this.layersMap = /* @__PURE__ */ new Map();
    this.resetLayersMap();
    this.generator = createGenerator();
  }
  getNodes(key) {
    const arrRef = this.layersMap.get(key);
    if (Array.isArray(arrRef)) {
      return arrRef;
    }
    const arr = [];
    this.layersMap.set(key, arr);
    return arr;
  }
  append(key, nodes) {
    const arr = this.getNodes(key);
    if (Array.isArray(nodes)) {
      arr.push(...nodes);
    } else {
      arr.push(nodes);
    }
  }
  resetLayersMap() {
    this.layersMap.set("base", []);
    this.layersMap.set("components", []);
    this.layersMap.set("utilities", []);
  }
}

function getDefaults() {
  return {
    atImportOptions: {},
    tailwindcssResolved: false,
    withOptions: true,
    withOptionsWalkCSSRuleObject(obj) {
      return obj;
    }
  };
}
function getClientOptions(opts) {
  return defu(opts, getDefaults());
}
function createContext(opts) {
  const ctx = new BaseContext();
  const options = getClientOptions(opts);
  ctx.options = options;
  return {
    ctx,
    getPlugins() {
      const opt = { ctx };
      const plugins = [atRulesRenamePlugin()];
      plugins.push(markLayerPlugin(opt));
      plugins.push(extractLayerPlugin(opt));
      options.postcssPlugins?.(plugins);
      return plugins;
    },
    async processString(rawCss) {
      const plugins = await this.getPlugins();
      const res = await postcss__default(plugins).process(rawCss, {
        from: void 0
      });
      return res;
    },
    generate: function() {
      const { generatorOptions } = options;
      return ctx.generator.generate(ctx, generatorOptions);
    }
  };
}

exports.createContext = createContext;
exports.getClientOptions = getClientOptions;
exports.getDefaults = getDefaults;
