import { Root, AcceptedPlugin, Node } from 'postcss';
import { GeneratorOptions } from '@babel/generator';
import { Options } from 'sass';
import { Config } from 'tailwindcss';
import { AtImportOptions } from 'postcss-import';
import { CSSRuleObject } from 'tailwindcss/types/config';

declare function createGenerator(): {
    generate: (ctx: BaseContext, opt?: GeneratorOptions) => string;
};

type LayerEnumType = 'base' | 'components' | 'utilities';

interface IProcessOptions {
    tailwindcssResolved?: boolean;
    tailwindcssConfig?: string | Config | {
        config: string | Config;
    } | undefined;
    sassOptions?: Options<'sync'>;
    atImportOptions?: AtImportOptions;
    generatorOptions?: GeneratorOptions;
    outSideLayerCss?: LayerEnumType;
    interceptors?: {
        css?: ((root: Root, ctx: BaseContext) => void)[];
    };
    withOptions?: boolean;
    withOptionsWalkCSSRuleObject?: (obj: CSSRuleObject, layer: '') => CSSRuleObject | CSSRuleObject[];
    postcssPlugins?: (plugins: AcceptedPlugin[]) => void;
}
type TailwindcssPluginOptions = {
    entries: string[];
    cacheDir?: string;
} & IProcessOptions;

declare class BaseContext {
    layersMap: Map<LayerEnumType, Node[]>;
    options: Required<TailwindcssPluginOptions> | undefined;
    generator: ReturnType<typeof createGenerator>;
    constructor();
    getNodes(key: LayerEnumType): Node[];
    append(key: LayerEnumType, nodes: Node | Node[]): void;
    resetLayersMap(): void;
}

export { BaseContext as B, type IProcessOptions as I, type TailwindcssPluginOptions as T };
