import fs from 'node:fs/promises';
import fs$1 from 'node:fs';
import path from 'node:path';
import postcss from 'postcss';
import atImport from 'postcss-import';
import * as t from '@babel/types';
import _babelGenerate from '@babel/generator';
import { pathToFileURL } from 'node:url';



// -- Unbuild CommonJS Shims --
import __cjs_url__ from 'url';
import __cjs_path__ from 'path';
import __cjs_mod__ from 'module';
const __filename = __cjs_url__.fileURLToPath(import.meta.url);
const __dirname = __cjs_path__.dirname(__filename);
const require = __cjs_mod__.createRequire(import.meta.url);
const markedLayerKey = Symbol("__tw_layer_name__");
const layerNodesKeys = ["base", "components", "utilities"];

const atRulesNameFilter = "twlayer";
const atRulesRenamePlugin = () => {
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-at-rule-rename-plugin",
    Once(root) {
      root.walkAtRules("layer", (rule) => {
        rule.name = atRulesNameFilter;
      });
    }
  };
};
atRulesRenamePlugin.postcss = true;
const markLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-mark-layer-plugin",
    Once(root) {
      root.walkAtRules(atRulesNameFilter, (rule) => {
        if (layerNodesKeys.includes(rule.params)) {
          const layerName = rule.params;
          for (const node of rule.nodes) {
            Object.defineProperty(node, markedLayerKey, {
              value: layerName,
              enumerable: false,
              configurable: true
            });
          }
          rule.parent?.insertBefore(rule, rule.nodes);
          rule.remove();
        }
      });
      if (ctx.options?.outSideLayerCss !== void 0) {
        root.walkRules((rule) => {
          const layerName = rule[markedLayerKey];
          if (!layerName) {
            Object.defineProperty(rule, markedLayerKey, {
              value: ctx.options?.outSideLayerCss,
              enumerable: false,
              configurable: true
            });
          }
        });
      }
    }
  };
};
markLayerPlugin.postcss = true;
const extractLayerPlugin = (options) => {
  const { ctx } = options;
  return {
    postcssPlugin: "postcss-css-to-tailwindcss-plugin-extract-layer-plugin",
    OnceExit(root) {
      root.walkRules((rule) => {
        const layerName = rule[markedLayerKey];
        if (layerName && layerNodesKeys.includes(layerName)) {
          ctx.append(layerName, rule);
        }
      });
      if (ctx.options?.interceptors?.css)
        for (const fn of ctx.options.interceptors.css) {
          fn(root, ctx);
        }
    }
  };
};
extractLayerPlugin.postcss = true;

function tryLoadSass() {
  return import('sass');
}
function tryLoadSassSync() {
  return require("sass");
}
function isExtSassFile(extname) {
  return extname === ".scss" || extname === ".sass";
}
async function sassCompile(path, sassOptions) {
  const { compile } = await tryLoadSass();
  const { css } = compile(path, sassOptions);
  return css;
}
function sassCompileSync(path, sassOptions) {
  const { compile } = tryLoadSassSync();
  const { css } = compile(path, sassOptions);
  return css;
}
async function sassCompileString(code, sassOptions) {
  const { compileString } = await tryLoadSass();
  const { css } = compileString(code, sassOptions);
  return css;
}

function _interopDefaultCompat(e) {
  return e && typeof e === "object" && "default" in e ? e.default : e;
}
const babelGenerate = _interopDefaultCompat(_babelGenerate);
const expandAST = ["addBase", "addComponents", "addUtilities", "theme"].map((x) => {
  return t.objectProperty(t.identifier(x), t.identifier("_" + x));
});
function expandAPI() {
  return expandAST;
}
function unwrapThemeFunctionArg(str) {
  if (str) {
    const start = str[0];
    const last = str.at(-1);
    if (start === '"' && last === '"' || start === "'" && last === "'") {
      return str.slice(1, -1);
    }
  }
  return str;
}
function makeObjectProperty(key, value) {
  return t.objectProperty(t.stringLiteral(key), value);
}
function addSuffix(value, important) {
  if (typeof value === "string") {
    return value + (important ? " !important" : "");
  }
  return value;
}
function makeObjectExpression(nodes) {
  const res = nodes.reduce((acc, _cur) => {
    if (_cur.type === "rule") {
      const cur = _cur;
      const key = cur.selector;
      const value = cur.nodes.reduce((cacc, ccur) => {
        if (ccur.type === "decl") {
          const v = ccur.value;
          const important = ccur.important;
          const prop = ccur.prop;
          const arr = [...v.matchAll(/theme\(([^()]+?)\)/g)];
          if (arr.length > 0) {
            const first = arr[0];
            if (first[0] === v) {
              const ccc = t.callExpression(t.identifier("_theme"), [t.stringLiteral(unwrapThemeFunctionArg(first[1]))]);
              cacc[prop] = important ? t.binaryExpression("+", ccc, t.stringLiteral(" !important")) : ccc;
            } else {
              let p = 0;
              const quasis = [];
              const expressions = [];
              for (let i = 0; i < arr.length; i++) {
                const hit = arr[i];
                if (typeof hit.index === "number") {
                  const raw = v.slice(p, hit.index);
                  p = hit.index + hit[0].length;
                  quasis.push(t.templateElement({ raw }));
                  expressions.push(t.callExpression(t.identifier("_theme"), [t.stringLiteral(unwrapThemeFunctionArg(hit[1]))]));
                  if (i === arr.length - 1) {
                    quasis.push(t.templateElement({ raw: addSuffix(v.slice(p), important) }));
                  }
                }
              }
              cacc[prop] = t.templateLiteral(quasis, expressions);
            }
          } else {
            cacc[prop] = t.stringLiteral(addSuffix(v, important));
          }
        }
        return cacc;
      }, {});
      if (acc[key]) {
        acc[key].push(
          ...Object.entries(value).map(([prop, value2]) => {
            return makeObjectProperty(prop, value2);
          })
        );
      } else {
        acc[key] = Object.entries(value).map(([prop, value2]) => {
          return makeObjectProperty(prop, value2);
        });
      }
    }
    return acc;
  }, {});
  return Object.entries(res).map(([key, value]) => {
    return t.objectProperty(
      t.stringLiteral(key),
      t.objectExpression(
        Object.values(
          value.reduce((acc, cur) => {
            if (cur.key.type === "StringLiteral") {
              acc[cur.key.value] = cur;
            }
            return acc;
          }, {})
        )
      )
    );
  });
}
function getVarName(layer) {
  return `_${layer}Css`;
}
const pluginNameMap = {
  base: "_addBase",
  components: "_addComponents",
  utilities: "_addUtilities"
};
function getFnName(key) {
  return pluginNameMap[key];
}
function getLegacyStatement(key, ctx) {
  return t.expressionStatement(t.callExpression(t.identifier(getFnName(key)), [t.objectExpression(makeObjectExpression(ctx.getNodes(key)))]));
}
const returnSelfArrowFnAst = t.arrowFunctionExpression([t.identifier("x")], t.identifier("x"));
const declReturnSelfArrowFnName = "returnSelfNoop";
const declReturnSelfArrowFnAst = t.variableDeclaration("const", [t.variableDeclarator(t.identifier(declReturnSelfArrowFnName), returnSelfArrowFnAst)]);
function createGenerator() {
  const callFnId = "_plugin";
  const pluginName = "css2TwPlugin";
  const withOptionsParamsId = "_options";
  const traverseMethodName = "withOptionsWalkCSSRuleObject";
  const requireStatement = t.variableDeclaration("const", [
    t.variableDeclarator(t.identifier(callFnId), t.callExpression(t.identifier("require"), [t.stringLiteral("tailwindcss/plugin")]))
  ]);
  const exportsStatement = t.expressionStatement(t.assignmentExpression("=", t.memberExpression(t.identifier("module"), t.identifier("exports")), t.identifier(pluginName)));
  function getLatestStatement(key, ctx) {
    const varName = getVarName(key);
    const fnName = getFnName(key);
    return [
      t.variableDeclaration("const", [
        t.variableDeclarator(
          t.identifier(varName),
          t.callExpression(t.identifier(traverseMethodName), [t.objectExpression(makeObjectExpression(ctx.getNodes(key))), t.stringLiteral(key)])
        )
      ]),
      t.expressionStatement(t.callExpression(t.identifier(fnName), [t.identifier(varName)]))
    ];
  }
  function generate(ctx, opt) {
    function getPluginBody() {
      const statements = ctx.options?.withOptions ? layerNodesKeys.flatMap((layer) => {
        return getLatestStatement(layer, ctx);
      }) : layerNodesKeys.map((layer) => {
        return getLegacyStatement(layer, ctx);
      });
      const innerPluginAst = t.functionExpression(
        null,
        [t.objectPattern(expandAPI())],
        t.blockStatement(statements)
      );
      if (ctx.options?.withOptions) {
        return t.callExpression(t.memberExpression(t.identifier(callFnId), t.identifier("withOptions")), [
          t.functionExpression(
            null,
            [t.assignmentPattern(t.identifier(withOptionsParamsId), t.objectExpression([]))],
            t.blockStatement([
              // const { withOptionsWalkCSSRuleObject = (x)=>x } = _options
              t.variableDeclaration("const", [
                t.variableDeclarator(
                  t.objectPattern([
                    t.objectProperty(t.identifier(traverseMethodName), t.assignmentPattern(t.identifier(traverseMethodName), t.identifier(declReturnSelfArrowFnName)))
                  ]),
                  t.identifier(withOptionsParamsId)
                )
              ]),
              t.returnStatement(innerPluginAst)
            ])
          ),
          t.functionExpression(null, [t.identifier(withOptionsParamsId)], t.blockStatement([t.returnStatement(t.objectExpression([]))]))
        ]);
      }
      return t.callExpression(t.identifier(callFnId), [innerPluginAst]);
    }
    const ast = t.file(
      t.program([requireStatement, declReturnSelfArrowFnAst, t.variableDeclaration("const", [t.variableDeclarator(t.identifier(pluginName), getPluginBody())]), exportsStatement])
    );
    const res = babelGenerate(ast, opt);
    return res.code;
  }
  return {
    generate
  };
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class BaseContext {
  constructor() {
    __publicField(this, "layersMap");
    __publicField(this, "options");
    __publicField(this, "generator");
    this.layersMap = /* @__PURE__ */ new Map();
    this.resetLayersMap();
    this.generator = createGenerator();
  }
  getNodes(key) {
    const arrRef = this.layersMap.get(key);
    if (Array.isArray(arrRef)) {
      return arrRef;
    }
    const arr = [];
    this.layersMap.set(key, arr);
    return arr;
  }
  append(key, nodes) {
    const arr = this.getNodes(key);
    if (Array.isArray(nodes)) {
      arr.push(...nodes);
    } else {
      arr.push(nodes);
    }
  }
  resetLayersMap() {
    this.layersMap.set("base", []);
    this.layersMap.set("components", []);
    this.layersMap.set("utilities", []);
  }
}

function isPlainObject(value) {
  if (value === null || typeof value !== "object") {
    return false;
  }
  const prototype = Object.getPrototypeOf(value);
  if (prototype !== null && prototype !== Object.prototype && Object.getPrototypeOf(prototype) !== null) {
    return false;
  }
  if (Symbol.iterator in value) {
    return false;
  }
  if (Symbol.toStringTag in value) {
    return Object.prototype.toString.call(value) === "[object Module]";
  }
  return true;
}

function _defu(baseObject, defaults, namespace = ".", merger) {
  if (!isPlainObject(defaults)) {
    return _defu(baseObject, {}, namespace, merger);
  }
  const object = Object.assign({}, defaults);
  for (const key in baseObject) {
    if (key === "__proto__" || key === "constructor") {
      continue;
    }
    const value = baseObject[key];
    if (value === null || value === void 0) {
      continue;
    }
    if (merger && merger(object, key, value, namespace)) {
      continue;
    }
    if (Array.isArray(value) && Array.isArray(object[key])) {
      object[key] = [...value, ...object[key]];
    } else if (isPlainObject(value) && isPlainObject(object[key])) {
      object[key] = _defu(
        value,
        object[key],
        (namespace ? `${namespace}.` : "") + key.toString(),
        merger
      );
    } else {
      object[key] = value;
    }
  }
  return object;
}
function createDefu(merger) {
  return (...arguments_) => (
    // eslint-disable-next-line unicorn/no-array-reduce
    arguments_.reduce((p, c) => _defu(p, c, "", merger), {})
  );
}
const defu = createDefu();

function getDefaultCacheDir() {
  return path.resolve(process.cwd(), "node_modules", ".css-to-tailwindcss-plugin");
}
function getDefaults() {
  const pkgPath = require.resolve("tailwindcss/package.json");
  const nodeModulesPath = path.dirname(path.dirname(pkgPath));
  return {
    atImportOptions: {},
    sassOptions: {
      importers: [
        {
          // https://github.com/webpack-contrib/sass-loader/blob/master/src/index.js#L55
          findFileUrl(url) {
            const twUrl = pathToFileURL(path.resolve(nodeModulesPath, url));
            return new URL(twUrl);
          }
        }
      ]
    },
    tailwindcssResolved: false,
    withOptions: true,
    withOptionsWalkCSSRuleObject(obj) {
      return obj;
    },
    cacheDir: getDefaultCacheDir()
  };
}
function getOptions(opts) {
  return defu(opts, getDefaults());
}

function createContext(opts) {
  const ctx = new BaseContext();
  const options = getOptions(opts);
  ctx.options = options;
  return {
    ctx,
    /**
     * @description Not recommended
     */
    getPluginsSync() {
      const opt = { ctx };
      const plugins = [atRulesRenamePlugin()];
      plugins.push(markLayerPlugin(opt));
      plugins.push(extractLayerPlugin(opt));
      options.postcssPlugins?.(plugins);
      return plugins;
    },
    async getPlugins() {
      const { tailwindcssConfig, tailwindcssResolved } = options;
      const opt = { ctx };
      const plugins = [atRulesRenamePlugin()];
      if (tailwindcssResolved && tailwindcssConfig) {
        const { default: tailwindcss } = await import('tailwindcss');
        plugins.push(tailwindcss(tailwindcssConfig));
      }
      plugins.push(markLayerPlugin(opt));
      plugins.push(extractLayerPlugin(opt));
      options.postcssPlugins?.(plugins);
      return plugins;
    },
    async process(entry) {
      const { sassOptions, atImportOptions } = options;
      const plugins = await this.getPlugins();
      let css;
      const extname = path.extname(entry);
      if (isExtSassFile(extname)) {
        try {
          css = await sassCompile(entry, sassOptions);
        } catch (error) {
          throw error;
        }
      } else {
        css = await fs.readFile(entry, "utf8");
        atImportOptions.root = path.dirname(entry);
        plugins.unshift(atImport(atImportOptions));
      }
      const res = await postcss(plugins).process(css, {
        from: void 0
      }).async();
      return res;
    },
    /**
     * @description Not recommended
     */
    processSync(entry) {
      const { sassOptions } = options;
      const plugins = this.getPluginsSync();
      let css;
      const extname = path.extname(entry);
      if (isExtSassFile(extname)) {
        try {
          css = sassCompileSync(entry, sassOptions);
        } catch (error) {
          throw error;
        }
      } else {
        css = fs$1.readFileSync(entry, "utf8");
      }
      const res = postcss(plugins).process(css).sync();
      return res;
    },
    /**
     * @description Not recommended
     */
    async processString(rawCss) {
      const { sassOptions } = options;
      const plugins = await this.getPlugins();
      const css = await sassCompileString(rawCss, sassOptions);
      const res = await postcss(plugins).process(css, {
        from: void 0
      }).async();
      return res;
    },
    generate: function() {
      const { generatorOptions } = options;
      return ctx.generator.generate(ctx, generatorOptions);
    }
  };
}

export { createContext as c, getOptions as g };
